/* GSKrabClient.m - this file is part of GSKrab
 *
 * Copyright (C) 2006 Wolfgang Sourdeau
 *
 * Author: Wolfgang Sourdeau <Wolfgang@Contre.COM>
 *
 * This file is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This file is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.  If not, write to
 * the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#import <Foundation/NSBundle.h>
#import <Foundation/NSConnection.h>
#import <Foundation/NSDistantObject.h>
#import <Foundation/NSException.h>
#import <Foundation/NSInvocation.h>
#import <Foundation/NSPort.h>
#import <Foundation/NSPortNameServer.h>
#import <Foundation/NSProcessInfo.h>
#import <Foundation/NSString.h>
#import <Foundation/NSTask.h>
#import <Foundation/NSThread.h>

#import "GSKrabServer.h"

#import "GSKrabClient.h"

@implementation GSKrabClient

+ (GSKrabClient *) sharedKrabClient
{
  static GSKrabClient *sharedKrabClient = nil;

  if (!sharedKrabClient)
    sharedKrabClient = [self new];

  fflush (stdout);

  return sharedKrabClient;
}

- (id) init
{
  if ((self = [super init]))
    {
      server = nil;
      serverNotFound = NO;
      clientName = nil;
      identifier = nil;
      actions = NSCreateMapTable (NSIntMapKeyCallBacks,
                                  NSObjectMapValueCallBacks,
                                  GSKrabMaxNumberOfKeys);
    }

  return self;
}

- (BOOL) _launchServer
{
  BOOL result;
  NSTask *serverTask;

  result = NO;
  NS_DURING
    {
      serverTask = [NSTask launchedTaskWithLaunchPath: @"opentool"
                           arguments: [NSArray arrayWithObject: @"gsks"]];
      while ([serverTask isRunning])
        [NSThread sleepUntilDate:
                    [NSDate dateWithTimeIntervalSinceNow: 0.01]];
      result = ([serverTask terminationStatus] == 0);
    }
  NS_HANDLER
    {
      if (![[localException name] isEqualToString: NSInvalidArgumentException])
        [localException raise];
    }
  NS_ENDHANDLER

  return result;
}

- (NSConnection *) _setupServiceConnectionWithName: (NSString *) serviceName
{
  NSConnection *connection;
  NSPort *port;

  port = [NSPort new];
  connection = [NSConnection connectionWithReceivePort: port
                             sendPort: nil];
  if ([[NSPortNameServer systemDefaultPortNameServer] registerPort: port
                                                      forName: serviceName])
    {
      [connection setRootObject: self];
      [connection setReplyTimeout: 1.0];
      [connection retain];
    }
  else
    connection = nil;

  [port release];

  return connection;
}

- (BOOL) _initialize
{
  NSString *application;

  application = [[NSBundle mainBundle] bundlePath];
  clientName = [NSString stringWithFormat: @"%@-GSKrabClient",
                         [application lastPathComponent]];

  server
    = [NSConnection rootProxyForConnectionWithRegisteredName: @"GSKrabServer"
                    host: nil];
  if (!server)
    {
      if ([self _launchServer])
        server = [NSConnection rootProxyForConnectionWithRegisteredName: @"GSKrabServer"
                               host: nil];
      else
        serverNotFound = YES;
    }

  if (server)
    {
      serviceConnection = [self _setupServiceConnectionWithName: clientName];
      [server setProtocolForProxy: @protocol (GSKrabServer)];
      if (!initialized)
        {
          identifier = [server registerAsClient: clientName
                               fromApplication: application];
          [identifier retain];
        }
      initialized = YES;
    }

  return (server != nil);
}

- (void) registerAction: (SEL) action
               onTarget: (id) target
             forKeyCode: (GSKrabKeyCode) krabKeyCode
{
  NSInvocation *keyAction;

  if (!serverNotFound && (server || [self _initialize]))
    {
      keyAction = [[NSInvocation alloc] initWithTarget: target
                                        selector: action];
      NSMapInsert (actions, (void *) krabKeyCode, keyAction);
      [server subscribeToKeyCode: krabKeyCode
              withIdentifier: identifier];
      [[server connectionForProxy] release];
      server = nil;
    }
}

- (void) performActionForKey: (GSKrabKeyCode) krabKeyCode
{
  NSInvocation *keyAction;

  keyAction = NSMapGet (actions, (void *) krabKeyCode);
  if (keyAction
      && [[keyAction target] respondsToSelector: [keyAction selector]])
    {
      [keyAction setArgument: &krabKeyCode
                 atIndex: 2];
      [keyAction invoke];
    }
}

- (void) dealloc
{
  NSArray *keyActions;

  [serviceConnection release];
  keyActions = NSAllMapTableValues (actions);
  [keyActions makeObjectsPerformSelector: @selector (release)];
  NSFreeMapTable (actions);
  [identifier release];

  [super dealloc];
}

@end
